#!/usr/bin/env python
"""
Gisaf job scheduler, orchestrating the background tasks
like remote device data collection, etc.
"""

import logging
from contextlib import asynccontextmanager

from fastapi.middleware.cors import CORSMiddleware

from gisaf.config import conf
from gisaf.redis_tools import store, shutdown_redis
from gisaf.scheduler import GSFastAPI, js, startup, Settings
from gisaf.scheduler_web import app as sched_app


formatter = logging.Formatter(
    "%(asctime)s:%(levelname)s:%(name)s:%(message)s", "%Y-%m-%d %H:%M:%S"
)
for handler in logging.root.handlers:
    handler.setFormatter(formatter)

logging.basicConfig(level=conf.gisaf.debugLevel)
logger = logging.getLogger("gisaf.scheduler_application")


@asynccontextmanager
async def lifespan(app: GSFastAPI):
    """
    Handle startup and shutdown: setup scheduler, etc
    """
    ## Startup
    js.start()
    await startup(settings)
    await store.setup(with_registry=False)
    yield
    await shutdown_redis()
    ## Shutdown
    pass


settings = Settings()
app = GSFastAPI(
    title=settings.app_name,
    lifespan=lifespan,
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.mount("/_sched", sched_app)
app.mount("/sched", sched_app)
